#!/usr/bin/env python3

import unittest

from gpuctl import PciDev, GpuDev

__all__ = ['GpuOk', 'GpuNak']

RATE_1M  = 1000
RATE_30M = 30000


class GpuOk(GpuDev):
    def is_gpu(self):
        return True

    def is_working(self):
        return True

    def get_temperature(self):
        return 30

    def set_speed(self, percentage):
        self.speed = percentage

class GpuNak(GpuDev):
    def is_gpu(self):
        return True

    def is_working(self):
        return True

    def get_temperature(self):
        return 70

    def set_speed(self, percentage):
        self.speed = percentage




#!/usr/bin/env python3

import sys
import signal
import argparse
import logging

from gpuctl import DRYRUN, GpuCtl, logger
from gpuctl import PciDev, GpuDev, GpuAMD, GpuNV

from gpuctl import MinerCtl, scan_miner

def show_net_stats():
    cmd = "lsof -i -P -n | grep -E 'miner|Miner' | grep LISTEN"

def run():

    gpu_ctl = None

    def sigstop(a, b):
        gpu_ctl.stop()
        # print(f'exit')
        # sys.exit(0)

    signal.signal(signal.SIGINT, sigstop)

    parser = argparse.ArgumentParser()

    # device
    parser.add_argument('-l', '--list', action='store_true',
                        help="list all GPU cards")
    parser.add_argument(
        '-s', '--slots', type=str, help="use PCI slot name to locate GPU (ie. 0000:01:00.0/0000:01:00.1)")
    parser.add_argument('-a', '--amd', action='store_true',
                        help="only use AMD GPU")
    parser.add_argument('-n', '--nvidia', action='store_true',
                        help="only use Nvidia GPU")

    # timer
    parser.add_argument('--interval', type=int, default=GpuCtl.INTERVAL,
                        help="monitoring interval")

    # fan control
    parser.add_argument('--set-speed', type=int, choices=range(0,101), metavar="[0-100]", 
                        help="set the fan speed (0~100)")
    parser.add_argument('-f', '--fan', type=int,
                        help="if temperature is exceed than FAN once, activate fan control (default:70)")
    parser.add_argument('-d', '--delta', type=int, default=2,
                        help="set fan speed if temperature diff %% is over DELTA (defaut:2)")

    # gpu lost failure detection
    parser.add_argument('--las', type=str, help="gpu lost action script")

    # temperature monitoring/actions
    parser.add_argument('--temp', type=int,
                        help="over temperature action threshold")
    parser.add_argument('--temp-cdown', type=int, default=GpuCtl.TEMP_CDOWN,
                        help="over temperature count down")
    parser.add_argument('--tas', type=str,
                        help="over temperature action script")

    # rate monitoring/actions
    parser.add_argument('--rms', type=str, help="rate monitoring script")
    parser.add_argument('--rate', type=int, default=1000,
                        help="under rate threshold (default: 1000 kh)")
    parser.add_argument('--rate-cdown', type=int, default=GpuCtl.RATE_CDOWN,
                        help="under rate count down")
    parser.add_argument('--ras', type=str, help="under rate action script")

    parser.add_argument('--curve', type=str,
                        help="set temp/fan-speed curve (ie. 0:0/10:10/80:100)")

    # remote management
    parser.add_argument('--scan', action='store_true',
                        help="scan miner's info through network management api")

    # misc
    parser.add_argument('-v', '--verbose',
                        action='store_true', help="show debug message")

    # parse arguments
    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    if args.scan:
        # 12, 8, 4, 4, 
        ports = scan_miner()
        for addr, port in ports:
            minerctl = MinerCtl(addr, port)
            r = minerctl.get_stats()
            if r == None:
                continue
            print(f"Miner : {r['miner']:12}")
            print(f"Uptime: {r['uptime']}s")
            print(f"Rate(kh) Temp Fan  ")
            print(f"======== ==== ==== ")
            for i in range(len(r['temp'])):
                print(f"{r['rate'][i]:8} {r['temp'][i]:3}c {r['fan'][i]:3}%")
            print('')
                
        sys.exit(0)

    # parse curve
    curve = None
    if args.curve:
        clst = args.curve.split('/')
        for c in clst:
            if ':' not in c:
                print(f'Invaid curve: [{args.curve}]')
                sys.exit(0)

        if clst:
            curve = [[int(c.split(':')[0]), int(c.split(':')[1])]
                     for c in clst]
        if curve and GpuDev.check_curve(curve):
            print(f'Applying curve: [{args.curve}]')
        else:
            print(f'Invaid curve: [{args.curve}]')

    # by slots
    gpu_devices = []
    slot_names = []

    # parse slot
    slots = None
    if args.slots:
        # by slot
        slots = args.slots.split('/')
        for sn in slots:
            sn = sn.strip().lstrip()
            sn = sn.strip('\'').lstrip('\'')
            pdev = PciDev.create(sn)
            gpu_dev = None
            if pdev and pdev.is_amd():
                gpu_dev = GpuAMD(pdev)
            if pdev and pdev.is_nvidia():
                gpu_dev = GpuNV(pdev)
            if gpu_dev and gpu_dev.is_gpu():
                gpu_devices.append(gpu_dev)
                slot_names.append(gpu_dev.pci_dev.slot_name)
    else:
        # by vendors
        vendors = []
        if args.amd:
            vendors.append('AMD')

        if args.nvidia:
            vendors.append('NVIDIA')

        pci_devices = PciDev.discovery(vendors)
        for pdev in pci_devices:
            gpu_dev = None
            if pdev and pdev.is_amd():
                gpu_dev = GpuAMD(pdev)
            if pdev and pdev.is_nvidia():
                gpu_dev = GpuNV(pdev)

            # remove duplicate gpu
            if gpu_dev and gpu_dev.is_gpu() and pdev.slot_name not in slot_names:
                gpu_devices.append(gpu_dev)

    # list monitored devices
    print("\n")
    print("ID Slot Name    Vendor   PCI-ID      Temp. Fan  PWR    Working")
    print("-- ------------ -------- ----------- ----- ---- ------ -------")

    cnt = 1
    for gpu in gpu_devices:
        pdev = gpu.pci_dev
        # set fan speed
        if args.set_speed != None:
            gpu.set_speed(args.set_speed)
        working = gpu.is_working()
        msg  = f"{cnt:2} {pdev.slot_name} {pdev.vendor_name():8} [{pdev.vendor_id}:{pdev.device_id}]"
        msg += f"{gpu.get_temperature():4}c {gpu.get_speed():3}% {gpu.get_pwr():6.2f}w {working}"
        print(msg)
        cnt += 1

    if args.list:
        sys.exit(0)

    if args.set_speed != None:
        sys.exit(0)

    if len(gpu_devices) == 0:
        print('No GPU found, abort !\n')
        sys.exit(0)

    # remove not working devices
    for gpu in list(gpu_devices):
        if gpu.is_working() == False:
            gpu_devices.remove(gpu)

    gpu_ctl = GpuCtl(gpu_devices=gpu_devices, fan=args.fan,
                     delta=args.delta, 
                     las=args.las,
                     temp=args.temp, tas=args.tas,
                     rms=args.rms, rate=args.rate, ras=args.ras, curve=curve)

    if not gpu_ctl.set_interval(intvl=args.interval, temp=args.temp_cdown, rate=args.rate_cdown):
        print(
            f'Interval error {args.interval}/{args.temp_cdown}/{args.rate_cdown} !\n')
        sys.exit(0)

    print(f"\ngpuctl: started\n")
    gpu_ctl.start()


if __name__ == '__main__':
    run()



#!/usr/bin/env python3

import pynvml as nv

from gpuctl import logger, DRYRUN
from gpuctl import FileVarMixn as fv
from gpuctl import ShellCmdMixin as sc

from . gpu_dev import PciDev, GpuDev

__all__ = ['GpuNV']


class GpuNV(GpuDev):

    CURVE = [[0, 0], [40, 0], [50, 0], [55, 50], [60, 60], [
        70, 70], [75, 80], [80, 100], [100, 100]]
    TEMP_DELTA = 2

    def __init__(self, pcidev):
        """
        check if the pci device is a GPU device and privided by vendor,
        if yes, fill in self.nvh
        """
        self.nvh = None
        self.display = ':0'
        self.gpu_flag = False

        INFO_PATH = "/proc/driver/nvidia/gpus/"
        super(GpuNV, self).__init__(pcidev)

        nv.nvmlInit()

        self.nvh = None
        self.name = 'NV '
        try:
            self.nvh = nv.nvmlDeviceGetHandleByPciBusId(
                pcidev.slot_name.encode())
            self.nv_id = nv.nvmlDeviceGetIndex(self.nvh)
            self.gpu_flag = True
            self.pcidev = pcidev
        except Exception as e:
            if e.value == nv.NVML_ERROR_GPU_IS_LOST:
                self.gpu_flag = True
            logger.debug(f'[{self.pci_dev.slot_name}/{self.name}] {str(e)}')

        self.curve = GpuNV.CURVE
        self.temp_delta = GpuNV.TEMP_DELTA

    def is_working(self):
        rv = True
        try:
            h = nv.nvmlDeviceGetHandleByPciBusId(
                self.pcidev.slot_name.encode())
        except Exception as e:
            rv = False
        return rv

    def is_gpu(self):
        return True if self.gpu_flag == True else False

    def set_speed(self, speed):

        if DRYRUN:
            return

        if self.nvh == None:
            return

        cmd = f"nvidia-settings -c {self.display} -a [gpu:{self.nv_id}]/GPUFanControlState=1 -a [fan:{self.nv_id}]/GPUTargetFanSpeed={speed}"
        logger.debug(f'exec: {cmd}')
        try:
            sc.exec_cmd(cmd)
            logger.debug(
                f'[{self.pci_dev.slot_name}/{self.name}] set speed {speed}%')
            self.speed = speed
        except:
            logger.error(f"{self.pci_dev.slot_name}/{self.name}] set fan speed failed !!")

    def get_speed(self):
        # NOTE: nvmlDeviceGetFanSpeed report wrong speed, use nvidia-settings instead
        # s = nv.nvmlDeviceGetFanSpeed(self.nvh)
        cmd = f"nvidia-settings -t -q  [fan:{self.nv_id}]/GPUTargetFanSpeed"
        logger.debug(f'exec: {cmd}')
        speed = 0
        try:
            s = sc.exec_cmd(cmd)
            speed = int(s)
            logger.debug(
                f'[{self.pci_dev.slot_name}/{self.name}] speed {speed}%')
        except:
            logger.error(f"{self.pci_dev.slot_name}/{self.name}] get fan speed failed !!")
        return speed

    def get_temperature(self):
        if self.nvh == None:
            return 0
        try:
            t = nv.nvmlDeviceGetTemperature(self.nvh, nv.NVML_TEMPERATURE_GPU)
            self.temperature = t
        except:
            logger.error(f"{self.pci_dev.slot_name}/{self.name}] get temperature failed !!")
            t = 0
        return t

    def get_pwr(self):
        if self.nvh == None:
            return 0
        try:
            pwr = nv.nvmlDeviceGetPowerUsage(self.nvh)/1000
            self.pwr = pwr
        except:
            logger.error(f"{self.pci_dev.slot_name}/{self.name}] get pwr failed !!")
            pwr = 0

        return pwr



#!/usr/bin/env python3

import sys
from gpuctl import logger, DRYRUN

import subprocess as sb

__all__ = ['FileVarMixn', 'ShellCmdMixin']

class FileVarMixn(object):
    @staticmethod
    def read_file_value(folder, filename):
        path = folder.rstrip('/') + '/' + filename
        value = None
        try:
            with open(path, 'r') as f:
                t = f.readline()
                value = int(t.strip('\n'))
        except:
            logger.error(f"cannot read value from {path} !!!")

        return value

    @staticmethod
    def read_file_str(folder, filename):
        path = folder.rstrip('/') + '/' + filename
        value = None
        try:
            with open(path, 'r') as f:
                t = f.readline()
                value = str(t.strip('\n'))
        except:
            logger.error(f"cannot read value from {path} !!!")

        return value

    @staticmethod
    def write_file_value(folder, filename, value):
        path = folder.rstrip('/') + '/' + filename
        value_str = str(value)

        if DRYRUN:
            return True
            
        rv = False
        try:
            with open(path, 'w') as f:
                f.write(value_str)
            rv = True
        except PermissionError:
            logger.error(f'needs root priviledge ({path}) !!!')
            sys.exit(0)

        except:
            logger.error(f"cannot write value {value} to {path} !!!")

        return rv

    @staticmethod
    def read_inifile_value(folder, filename, cfname):
        path = folder.rstrip('/') + '/' + filename
        value = None
        try:
            with open(path, 'r') as f:
                for line in f:
                    name, value = line.lstrip(' ').strip('\n').split('=', 1)
                    if name == cfname:
                        break
        except:
            pass
        return value

class ShellCmdMixin(object):
    @staticmethod
    def exec_cmd(cmd):
        session = sb.Popen(cmd, shell=True, stdout=sb.PIPE, stderr=sb.PIPE)
        data, err = session.communicate()
        if err:
            raise Exception("Error: " + str(err))
        return data

    @staticmethod
    def exec_script(script, params=None):
        """
        script must have executable permission (chmod +x script)
        """
        cmd = script + ' ' + params
        session = sb.Popen([cmd], shell=True, stdout=sb.PIPE, stderr=sb.PIPE)
        data, err = session.communicate()
        if err:
            raise Exception("Error: " + str(err))
        return data


#!/usr/bin/env python3

import sys
import time
import signal
import socket
import argparse
import json

from gpuctl import ShellCmdMixin as sc

__all__ = ['MinerCtl', 'scan_miner']

HOST, PORT = "localhost", 3333

MAX_DATA = 512
RX_TIMEOUT = 2

UP_TIME_THRESHOLD = (3 * 60) # seconds

QUERY_INTERVAL = 5 # seconds
OVER_THRESHOLD_PERIOD = 180 # seconds

RATE_COUNTDOWN = int(OVER_THRESHOLD_PERIOD/QUERY_INTERVAL)
TEMP_COUNTDOWN = int(OVER_THRESHOLD_PERIOD/QUERY_INTERVAL)

class CmdType:
    MSG_GET_STAT1 = b'{"method": "miner_getstat1", "jsonrpc": "2.0", "id": 0 }\n'
    MSG_RESTART  = b'{"method":"miner_restart", "id":1,"jsonrpc":"2.0"}\n'
    MSG_REBOOT  = b'{"method":"miner_boot", "id":2,"jsonrpc":"2.0"}\n'

def parse_stats(stats, disp_flag=False):

    if stats == None:
        return None

    result = {}

    result['miner'] = stats[0]
    result['uptime'] = int(stats[1]) * 60

    summary = stats[2].split(';')
    result['total'] = int(summary[0])
    result['accept'] = int(summary[1])
    result['reject'] = int(summary[2])
    result['rate'] = [int(i) for i in stats[3].split(';')] 

    tf = stats[6].split(';')
    result['temp'] = tf[0::2]
    result['temp'] = [int(i) for i in result['temp']]
    result['temp_high'] = max(result['temp']) 
    result['fan'] = tf[1::2]
    result['fan'] = [int(i) for i in result['fan']] 
    
    if disp_flag:
        print(f"{result['miner']}: {result['uptime']}m {result['rate']}Mh/s {result['temp']}C {result['fan']}%")

    return result

def scan_miner():
    tcp_ports = []
    cmd = "lsof -i -P -n | grep -E 'miner|Phoenix' | grep LISTEN"
    ret = sc.exec_cmd(cmd)
    items = ret.split(b' ')
    for i in items:
        address = i.split(b':')
        if len(address) == 2: # address:port
            addr = 'localhost' if address[0] == '*' else address[0]
            tcp_ports.append((addr, int(address[1])))
    return tcp_ports

def do_cmd(host, port, cmd):
    # Create a TCP/IP socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # Connect the socket to the port where the server is listening
    server_address = ('localhost', port)
    json_data = None

    try:
        sock.connect(server_address)

        # Send data
        message = cmd
        sock.sendall(message)

        # Look for the response
        sock.settimeout(RX_TIMEOUT)
        data = sock.recv(MAX_DATA)
        if data:
            json_data = json.loads(data)['result']

    # except socket.timeout as e:
    except:
        print('timeout')
    finally:
        # print('closing socket')
        sock.close()

    return json_data

class MinerCtl:
    def __init__(self, address, port):
        self.address = address
        self.port = port

    def get_stats(self):
        result = do_cmd(self.address, self.port, CmdType.MSG_GET_STAT1)
        stats = parse_stats(result)
        return stats

    def restart(self):
        result = False
        ret = do_cmd(self.address, self.port, CmdType.MSG_RESTART)
        if ret != None:
            result = ret
        return result

    def reboot(self):
        result = False
        ret = do_cmd(self.address, self.port, CmdType.MSG_REBOOT)
        if ret != None:
            result = ret
        return result

def monitor(mctl, rate, temp, verbose):

    rate_cnt = RATE_COUNTDOWN
    temp_cnt = TEMP_COUNTDOWN

    print('start monitoring ....')
    while True:
        time.sleep(QUERY_INTERVAL)
        result = mctl.get_stats()
        if result == None:
            print('cannot get stats')
            continue
        
        if int(result['uptime']) < UP_TIME_THRESHOLD:
            print(f"wait util {UP_TIME_THRESHOLD}, uptime: {result['uptime']} ...")
            rate_cnt = RATE_COUNTDOWN
            temp_cnt = TEMP_COUNTDOWN
            continue

        if rate != None and result['total'] <= rate:
            rate_cnt -= 1
            print(f"rate under threshold current {result['total']} kh/s threshold {rate} kh/s countdown {rate_cnt}/{RATE_COUNTDOWN}")
        else:
            rate_cnt = RATE_COUNTDOWN

        if temp != None and result['temp_high'] >= temp:
            temp_cnt -= 1
            print(f"temp over threshold current {result['temp_high']}c threshold {temp}c countdown {temp_cnt}/{TEMP_COUNTDOWN}")
        else:
            temp_cnt = TEMP_COUNTDOWN

        if verbose:
            print(f"rate: {result['total']} kh/s temp: {result['temp_high']}c countdown {rate_cnt}/{temp_cnt}")

        if rate_cnt <= 0 or temp_cnt <= 0:
            print(f"restart miner: rate {result['total']} kh/s temp {result['temp_high']}c")
            mctl.restart()
            rate_cnt = RATE_COUNTDOWN
            temp_cnt = TEMP_COUNTDOWN
            print(f"wait 30s for miner restart") 
            time.sleep(30)

if __name__ == "__main__":

    def sigstop(a ,b):
        print(f'exit') 
        sys.exit(0)

    signal.signal(signal.SIGINT, sigstop)

    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str, default=HOST, help="host address (default: localhost)")
    parser.add_argument('--port', type=int, default=PORT, help="tcp port number (default: 3333)")
    parser.add_argument('--cmd', type=int, default=0, help="0: get_stats, 1: restart (default: 0)")
    parser.add_argument('--rate', type=int, help="under rate restart")
    parser.add_argument('--temp', type=int, help="over temp restart")
    parser.add_argument('--wait', type=int, help="wait for rx timeout in seconds (default: 2)")
    parser.add_argument('--scan', action='store_true',
                        help="scan miner's info through network management api")
    parser.add_argument('--verbose', action='store_true', help="show debug message")
    args = parser.parse_args()

    # HOST = (args.host or HOST)
    # PORT = (args.port or PORT)

    if args.scan:
        # 12, 8, 4, 4, 
        ports = scan_miner()
        for addr, port in ports:
            minerctl = MinerCtl(addr, port)
            r = minerctl.get_stats()
            if r == None:
                continue
            print(f"Miner : {r['miner']:12}")
            print(f"Uptime: {r['uptime']}s")
            print(f"Rate(kh) Temp Fan  ")
            print(f"======== ==== ==== ")
            for i in range(len(r['temp'])):
                print(f"{r['rate'][i]:8} {r['temp'][i]:3}c {r['fan'][i]:3}%")
            print('')
                
        sys.exit(0)
        

    minerctl = MinerCtl(args.host, args.port)

    result = None
    if args.rate != None or args.temp != None:
        monitor(minerctl, args.rate, args.temp, args.verbose)
    else:
        if args.cmd == None or args.cmd == 0:
            result = minerctl.get_stats()
        elif args.cmd == 1:
            result = minerctl.restart()
        elif args.cmd == 2:
            result = minerctl.reboot()
        else:
            sys.exit(0)
    
    print('result: ', result)


#!/usr/bin/env python3

import os

from gpuctl import logger, DRYRUN
from gpuctl import FileVarMixn as fv
from gpuctl import ShellCmdMixin as sc

from . gpu_dev import PciDev, GpuDev

__all__ = ['GpuAMD']


class GpuAMD(GpuDev):
    DRM_DIR = '/sys/class/drm/'
    PWM_MAX = 'pwm1_max'
    PWM_MIN = 'pwm1_min'
    PWM_MODE = 'pwm1_enable'
    PWM_VAL = 'pwm1'
    TEMP_VAL = 'temp1_input'
    PWR_VAL = 'power1_average'

    CURVE = [[0, 0], [40, 0], [50, 30], [55, 50], [60, 60], [
        70, 70], [75, 80], [80, 100], [100, 100]]
    TEMP_DELTA = 2

    def __init__(self, pcidev):
        """
        ls /sys/class/drm | grep "^card[[:digit:]]$"
        check if the pci device is a GPU device and privided by vendor,
        if yes, fill in self.card
        """

        self.card = None
        super(GpuAMD, self).__init__(pcidev)
        card_dirs = [c for c in os.listdir(
            GpuAMD.DRM_DIR) if len(c) <= 6]  # card##

        # locate card dir
        self.hwmon_dir = None
        for card in card_dirs:
            ue_file = f'{GpuAMD.DRM_DIR}{card}/device/uevent'
            slot_name = fv.read_inifile_value('', ue_file, 'PCI_SLOT_NAME')
            if slot_name == self.pci_dev.slot_name:
                self.name = 'AMD'
                self.card = card
                # locate hwmon# dir
                hwmon_root = f'{GpuAMD.DRM_DIR}{self.card}/device/hwmon'
                t = os.listdir(hwmon_root)
                self.hwmon_dir = hwmon_root + '/' + t[0]
                break
        
        # fan
        if self.card:
            self.max = fv.read_file_value(self.hwmon_dir, GpuAMD.PWM_MAX)
            self.min = fv.read_file_value(self.hwmon_dir, GpuAMD.PWM_MIN)

        # curve
        self.curve = GpuAMD.CURVE
        self.temp_delta = GpuAMD.TEMP_DELTA

    def is_gpu(self):
        return True if self.card else False

    def is_working(self):
        return True
        
    def set_speed(self, speed):
        """
        speed: 0~100 (percentage)
        """
        # hwmon_dir = f'{GpuAMD.DRM_DIR}{self.card}/device/hwmon/hwmon?/'
        speed = 100 if speed > 100 else speed
        speed = 0 if speed < 0 else speed
        pwm = int((self.max - self.min) * speed / 100)
        fv.write_file_value(self.hwmon_dir, GpuAMD.PWM_MODE, 0)  # 0: disabled
        fv.write_file_value(self.hwmon_dir, GpuAMD.PWM_MODE, 1)  # 1: manual
        fv.write_file_value(self.hwmon_dir, GpuAMD.PWM_VAL, pwm)
        self.speed = speed
        logger.debug(f'[{self.pci_dev.slot_name}/{self.name}] set speed {speed}% pwm {pwm}')

    def get_speed(self):
        pwn = fv.read_file_value(self.hwmon_dir, GpuAMD.PWM_VAL)
        speed = int(pwn/(self.max-self.min) * 100)
        return speed

    def get_temperature(self):
        # hwmon_dir = f'{GpuAMD.DRM_DIR}{self.card}/device/hwmon/hwmon?/'
        temp = int(fv.read_file_value(self.hwmon_dir, GpuAMD.TEMP_VAL) / 1000)
        self.temperature = temp
        return temp

    def get_pwr(self):
        pwr = int(fv.read_file_value(self.hwmon_dir, GpuAMD.PWR_VAL) / 1000000)
        self.pwr = pwr
        return pwr




#!/usr/bin/env python3

import os
import sys
import logging

logger = logging.getLogger(__name__)
DRYRUN = bool(os.environ.get('DRYRUN', False))

logging.basicConfig(stream=sys.stdout,
                    format='%(asctime)s %(levelname)-8s %(message)s',
                    datefmt='%H:%M:%S',
                    level=logging.INFO)

__all__ = []

from . utils import *
__all__ += utils.__all__

from . gpu_dev import *
__all__ += gpu_dev.__all__

from . gpu_amd import *
__all__ += gpu_amd.__all__

from . gpu_nv import *
__all__ += gpu_nv.__all__

from . gpu_mock import *
__all__ += gpu_mock.__all__

from . gpu_ctl import *
__all__ += gpu_ctl.__all__

from . miner_ctl import *
__all__ += gpu_ctl.__all__



#!/usr/bin/env python3

import os
import sys
import time
import re
import argparse
import subprocess as sb
from threading import Thread
import pynvml as nv

from gpuctl import logger
from gpuctl import PciDev, GpuAMD, GpuNV
from gpuctl import FileVarMixn as fv
from gpuctl import ShellCmdMixin as sc

__all__ = ['GpuCtl']


class GpuCtl():

    INTERVAL = 1
    TEMP_CDOWN = 120 # seconds
    RATE_CDOWN = 120 # seconds

    def __init__(self, **kwargs):

        self.vendors = None

        # overwrite default value with arguments
        valid_keys = ["slots", "gpu_devices",
                      "fan", "delta", "temp", "tas", "rms", "rate", "ras", "las", "curve"]
        for key in valid_keys:
            setattr(self, key, kwargs.get(key))

        self.thread = None
        self.abort = False

        # if curve is set, apply curve to all GPUs
        if self.curve:
            for gpu_dev in self.gpu_devices:
                gpu_dev.set_curve(self.curve)
            
        # fan
        self.fan_ctrl_flag = False

        # action script
        self.interval = GpuCtl.INTERVAL
        self.temp_cdown = GpuCtl.TEMP_CDOWN
        self.rate_cdown = GpuCtl.RATE_CDOWN

        for gpu in self.gpu_devices:
            gpu.prev_temp = None
            gpu.temp_cdown = self.temp_cdown
            gpu.rate_cdown = self.rate_cdown

    def _fan_control(self, gpu, t):
        if t and not self.fan_ctrl_flag and (self.fan and t > self.fan):
            logger.debug('fan control is activated')
            self.fan_ctrl_flag = True

        if not self.fan_ctrl_flag:
            return

        speed = gpu.temp_to_speed(t)
        ptemp = gpu.prev_temp
        # if pwm change more than delta percentage, then do action
        if (ptemp == None or abs(t - ptemp) > gpu.temp_delta):
            logger.info(
                f"[{gpu.pci_dev.slot_name}/{gpu.name}] current temp. {t}c set speed {speed}%")
            gpu.set_speed(speed)
            gpu.prev_temp = t

    def _failure_action(self, gpu):
        if gpu.is_working() == False:
            if self.las:
                try:
                    rv = sc.exec_script(self.las, params=gpu.pci_dev.slot_name)
                    logger.info(f"[{gpu.pci_dev.slot_name}/{gpu.name}] gpu failure, exec script {self.las}")
                    time.sleep(5) # wait 5s
                except:
                    logger.error(f"{gpu.pci_dev.slot_name}/{gpu.name}] gpu failure, exec script {self.las} failed !!")
            else:
                logger.warning(f"[{gpu.pci_dev.slot_name}/{gpu.name}] gpu failure, no action script defined")


    def _temp_action(self, gpu, t):
        if self.temp and t > self.temp:
            logger.warning(f"[{gpu.pci_dev.slot_name}/{gpu.name}] temp: {t}c/{self.temp}c CD: {gpu.temp_cdown}s")
            gpu.temp_cdown -= self.interval
        else:
            gpu.temp_cdown = self.temp_cdown
            return
        
        # take action
        if gpu.temp_cdown <= 0:
            if self.tas:
                try:
                    rv = sc.exec_script(self.tas, params=gpu.pci_dev.slot_name)
                    logger.info(f"[{gpu.pci_dev.slot_name}/{gpu.name}] over heat, exec script {self.tas}")
                    logger.info(f"[{gpu.pci_dev.slot_name}/{gpu.name}] result: {rv.decode('utf-8')}")
                except:
                    logger.error(f"{gpu.pci_dev.slot_name}/{gpu.name}] over heat, exec script {self.tas} failed !!")
            else:
                logger.warning(f"[{gpu.pci_dev.slot_name}/{gpu.name}] over heat, no action script defined")

            gpu.temp_cdown = self.temp_cdown

    def _get_hashrate(self, gpu, script):
        rate = None
        try:
            rv = sc.exec_script(script, params=gpu.pci_dev.slot_name)
            rate = int(rv.decode('utf-8').strip('\n'))
        except:
            logger.error(f"{gpu.pci_dev.slot_name}/{gpu.name}] get hashrate, exec script {script} failed !!")
        return rate

    def _rate_action(self, gpu, rate):
        if rate and self.rate and rate < self.rate:
            logger.warning(f"[{gpu.pci_dev.slot_name}/{gpu.name}] rate: {rate}/{self.rate} CD: {gpu.rate_cdown}s")
            gpu.rate_cdown -= self.interval
        else:
            gpu.rate_cdown = self.rate_cdown
            return
        
        # take action
        if gpu.rate_cdown <= 0:
            if self.ras:
                try:
                    rv = sc.exec_script(self.ras, params=gpu.pci_dev.slot_name)
                    logger.info(f"[{gpu.pci_dev.slot_name}/{gpu.name}] under rate, exec script {self.tas}")
                    logger.info(f"[{gpu.pci_dev.slot_name}/{gpu.name}] result: {rv.decode('utf-8')}")
                except:
                    logger.error(f"{gpu.name} under rate, exec script {self.tas} failed !!")
            else:
                logger.warning(f"[{gpu.pci_dev.slot_name}/{gpu.name}] under rate, no action script defined")

            gpu.rate_cdown = self.rate_cdown

    def _gpu_thread(self):
        while self.abort != True:
            for gpu in self.gpu_devices:

                time.sleep(self.interval)

                # gpu failure
                self._failure_action(gpu)

                t = gpu.get_temperature()
                if t == None:
                    continue

                # fan speed
                self._fan_control(gpu, t)

                # action scripts

                # temp
                self._temp_action(gpu, t)

                # rate
                rate = self._get_hashrate(gpu, self.rms) if self.rms else None
                self._rate_action(gpu, rate)

                temp_str = f'temp: {t}/{gpu.temp_cdown}s' if t else ''
                rate_str = f'rate: {rate}/{gpu.rate_cdown}s' if rate else ''
                msg = temp_str + ' ' + rate_str
                logger.debug(f"[{gpu.pci_dev.slot_name}/{gpu.name}] {msg}")


    def add_gpu_devices(self, gpu_devices):
        cnt = 0
        for gpu in gpu_devices:
            dup_flag = False
            for g in self.gpu_devices:
                if gpu.pci_dev.slot_name == g.pci_dev.slot_name:
                    dup_flag = True
                    logger.debug(f'duplicate slot {gpu.pci_dev.slot_name}')
                    break
            if not dup_flag:
                cnt += 1
                self.gpu_devices.append()
        return cnt

    def start(self):
        if len(self.gpu_devices) == 0:
            logger.error('No GPU found !!!')
            return
        self.thread = Thread(target=self._gpu_thread)
        self.thread.start()

    def stop(self):
        logger.info("exit")
        self.abort = True
        self.thread.join()

    def set_interval(self, intvl=None, temp=None, rate=None):

        interval = intvl if intvl else self.interval
        temp_cdown = temp if temp else self.temp_cdown
        rate_cdown = rate if rate else self.rate_cdown

        if interval <= 0 or temp_cdown < interval or rate_cdown < interval:
            logger.error(f'invalid intervals {intvl} {temp} {rate}!!!')
            return False

        self.interval = interval
        self.temp_cdown = temp_cdown
        self.rate_cdown = rate_cdown

        for gpu in self.gpu_devices:
            gpu.temp_cdown = self.temp_cdown
            gpu.rate_cdown = self.rate_cdown

        return True

if __name__ == '__main__':
    pci_devices = PciDev.discovery()
    gpu_devices = []
    for pdev in pci_devices:
        gpu_dev = None
        if pdev and pdev.is_amd():
            gpu_dev = GpuAMD(pdev)
        if pdev and pdev.is_nvidia():
            gpu_dev = GpuNV(pdev)

        # remove duplicate gpu
        if gpu_dev and gpu_dev.is_gpu():
            gpu_devices.append(gpu_dev)

    for gpu in gpu_devices:
        rate = gpu.get_hashrate('./scripts/rate.sh')
        print(f'[{gpu.pci_dev.slot_name}/{gpu.name}] rate: {rate}')


from setuptools import setup
from setuptools import find_packages

with open('README.md') as f:
    readme = f.read()

setup(
    name='gpuctl',
    version='0.3.4',
    packages=find_packages(),
    url='https://github.com/Ed-Yang/gpuctl',
    download_url='https://github.com/Ed-Yang/gpuctl/archive/refs/tags/v0.3.4.tar.gz',
    license='MIT',
    author='Edward Yang',
    author_email='edwardyangyang@hotmail.com',
    description='GPU contorl and failure notification',
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords='gpu amd nvidia ai pytorch tensorflow torch mining ethereum bitcoin cryptocurrency ethminer nsfminer phoenixminer',
    python_requires='>=3',
    install_requires=[
        'setuptools',
        'wheel',
        'pynvml'
    ],
    entry_points="""
        [console_scripts]
        gpuctl=gpuctl.main:run
    """,
)




#!/usr/bin/env python3

import os
import sys
import time
import re
import pynvml as nv

from gpuctl import logger
from gpuctl import FileVarMixn as fv
from gpuctl import ShellCmdMixin as sc

__all__ = ['PciDev', 'GpuDev']

class PCI_VID:
    VID_AMD = '1002'
    VID_NV = '10DE'

VENDOR_NAME  = {
    PCI_VID.VID_AMD: 'AMD', 
    PCI_VID.VID_NV: 'NVIDIA'
    }

class PciDev():
    """
    slot_name: DDDD:BB:SS.F (domain:bus:slot/device:function)
    """

    @staticmethod
    def parse_slot_name(slot_name):
        try:
            ids = slot_name.split(':')
            domain = ids[0]
            bus_id = ids[1].upper()
            slot_id = ids[2].split('.')[0].upper()
            function = ids[2].split('.')[1].upper()
            return domain, bus_id, slot_id, function
        except:
            # logger.error(f"Invaid slot name: {slot_name}")
            raise Exception(f"Invaid slot name: {slot_name}")



    def __init__(self, slot_name, pci_id, desc):
        self.desc = desc
        self.slot_name = slot_name

        try:
            self.domain, self.bus_id, self.slot_id, self.function = self.parse_slot_name(
                slot_name)
            self.vendor_id = pci_id.split(':')[0].upper()
            self.device_id = pci_id.split(':')[1].upper()
            self._vendor_name = VENDOR_NAME.get(self.vendor_id, 'Other')
        except:
            self.vendor_id = ''
            self.device_id = ''
            self._vendor_name = ''

    def vendor_name(self):
        return self._vendor_name

    def is_amd(self):
        if self.vendor_id == PCI_VID.VID_AMD:
            return True
        else:
            return False

    def is_nvidia(self):
        if self.vendor_id == PCI_VID.VID_NV:
            return True
        else:
            return False

    @classmethod
    def create(cls, slot_name):
        """
        /sys/bus/pci/devices/
        """
        slot_name = slot_name.strip('\'').rstrip('\'')
        SLOT_DIR = f'/sys/bus/pci/devices/{slot_name}'
        domain, bus_id, slot_id, function = cls.parse_slot_name(
            slot_name)

        try:
            vendor_id = fv.read_file_str(SLOT_DIR, 'vendor').strip('0x')
            device_id = fv.read_file_str(SLOT_DIR, 'device').strip('0x')
            pci_id = vendor_id.upper() + ':' + device_id.upper()
            desc = 'Other'
            if vendor_id == PCI_VID.VID_AMD:
                desc = 'AMD'
            elif vendor_id == PCI_VID.VID_NV:
                desc = 'Nvidia'
            else:
                desc = 'Other'
        except:
            return None

        return cls(slot_name, pci_id, desc)

    @staticmethod
    def discovery(vendor_filter=None):
        """
        cretiria: 
            - AMD or Nvida PCI devices
        """
        # self.gpu_devices = []
        pci_devices = []
        cmd = ''
        if vendor_filter:
            cmd = "lspci -D -nn | awk '/" f"{'|'.join(vendor_filter)}" + "/{print $0}'"
        else:
            cmd = "lspci -D -nn"
        try:
            data = sc.exec_cmd(cmd)
            lines = data.decode("utf-8").split('\n')
        except:
            logger.error(f'exec_cmd: {cmd} failed !!')
            lines = []

        desc = ''
        for line in lines:
            if line == '':
                continue
            slot_name, desc = line.split(' ', 1)
            group = re.findall('\[(.*?)\]', desc)
            pdev = None
            for g in group:
                if ':' in g:
                    pcidev = None
                    try:
                        pcidev = PciDev(slot_name, g.upper(), desc)
                        break
                    except:
                        pass

            if vendor_filter:
                if pcidev.vendor_name() in vendor_filter:
                    pci_devices.append(pcidev)
            else:
                pci_devices.append(pcidev)

        return pci_devices

    @staticmethod
    def discovery_by_slots(vendor, slots):
        pci_devices = []
        for sn in slots:
            sn = sn.strip().lstrip()
            sn = sn.strip('\'').lstrip('\'')
            pdev = PciDev.create(sn)

            if pdev != None:
                pci_devices.append(pdev)

class GpuDev():
    def __init__(self, pcidev=None):
        # devices
        self.name = 'Other'
        self.dev_info = {}
        self.dev_info['bios_ver'] = ""
        self.pci_dev = pcidev

        # curve
        self.curve = None
        self.sp_delta = None

        # current info
        self.speed = 0  # percentage
        self.temperature = 0
        pass

    # def get_dev_info(self):
    #     return self.dev_info

    def is_gpu(self):
        raise NotImplementedError

    def is_working(self):
        raise NotImplementedError

    def get_temperature(self):
        raise NotImplementedError

    def get_pwr(self):
        raise NotImplementedError

    def get_speed(self, percentage):
        raise NotImplementedError

    def set_speed(self, percentage):
        raise NotImplementedError

    @staticmethod
    def check_curve(curve):
        if len(curve) < 2:
            return False

        for i in range(len(curve)-1):
            if curve[i][0] > curve[i+1][0] or curve[i][1] > curve[i+1][1]:
                return False
        return True

    def set_curve(self, curve):
        self.curve = curve

    def temp_to_speed(self, temp):
        """
        assume temp is greater than 0
        """
        speed = None
        for t in range(len(self.curve)):
            if temp < self.curve[t][0]:
                speed = int(self.curve[t - 1][1] + (self.curve[t][1] - self.curve[t - 1][1]) / (
                    self.curve[t][0] - self.curve[t - 1][0]))
                break

        return speed

if __name__ == '__main__':

    pass
